// Time-stamp: </home/wrf/c/2018/connect2018/connect2.cc, Sat, 19 May 2018, 18:00:54 EDT, w540, wrf, W. Randolph Franklin, http://wrfranklin.org/>
//
// Find 6- or 26-connected components in a 3-D array of voxels.
//
// See the web page for how to use this program.
//
// If preprocessor symbol TIMEINPUTS is defined on compilation, then
// read the data but don't process it.  This lets us time reading the input.
//
//
//		      Wm. Randolph Franklin
//                       Associate Professor
//			 wrf@ecse.rpi.edu
//	      http://www.ecse.rpi.edu/Homepages/wrf/
//		  +1 (518) 276-6077;  Fax: -6261
//ECSE Dept., 6026 JEC, Rensselaer Polytechnic Inst, Troy NY, 12180 USA
//			 (PGP key available)
//


// #define TIMEINPUTS

// Print a status every 3 seconds if WATCHDOG is defined
#define WATCHDOG

// Use static instead of dynamically allocated arrays?
//#define STATIC

// INCLUDE FILES

#include "header"
#include <stdlib.h>
#include <errno.h>
#include <fstream>
#include <sstream>
#include <time.h>
#include <stdio.h>
#include <sys/time.h>

using namespace std;

// CONSTANTS

#ifdef STATIC
const int mx=1;
const int my=18574;          // Max size of volume.
const int mz=19111;

//const int mx=545;
//const int my=545;          // Max size of volume.
//const int mz=545;

//const int my=2000;          // Max size of volume.
//const int mz=2000;

const int mruns= 4000000;
const int mcomps= 1000000;		// Max number of connected components

#else
int mx, my, mz, mruns, mcomps;
#endif

const int msize=20000;  // Size of histogram of component sizes.


// STRUCTURES

// RUN: One run of empty voxels, with fixed (x,y), from zlo to zhi.
// As the runs are being joined into components, the runs in each
// component are linked into a tree.  ROOT<0 if this is the root run
// of a component tree.  In this case, it's the negative of the number
// of voxels in this tree. Else, ROOT is the run # of either the root
// run, or of another run that is closer to the root.

// Once the components have been formed, the root field is used to
// link each component's runs in a list.
  

class run	
{
public: 
  short int x, y, zlo, zhi;
  signed int root;     
};

class component
{
public:
  int nruns;          // # runs in each component.
  int runs;           // Pointer to first run of this component.
  int vol;            // Volume
  int area;           // Surface area.
};

//  VARIABLES

int errno;
int nx;
int ny;
int nz;
int debug;			// level of debugging output
int conn26=0;            //  Do 26-connectivity instead of 6-connectivity?
int key;            // Select nonstandard outputs.
int nruns;          // # runs
int ncomps;			// # connected components
int nvoxels;		// # 0 voxels
int nadjacent;		// # times 2 adjacent overlapping runs processed.
int nnonoverlap;	// # times a pair of possibly adjacent runs tested,
                        // but they didn't overlap.
int nfollrptr; // # times a pointer was followed when connecting two runs.
int nfindroot=0;   // # times find_root_of_run is called.
int lfindroot=0;   // # pointers followed in find_root_of_run. If the run is 
                    // the root, then 0 is added here.
int nupdateroot=0;   // # times update_root is called.
int lupdateroot=0;   // # pointers followed in update_root.
int nsametree;	// # times two adjacent runs were already in the same component.
int nfindrootslink;  // # times find_roots find a run not pointing directly to its root. 

// ROWS is conceptually a 2-D array, indexed rows[x][y], containing the 
// number (the index in runs) of the highest run in this row.   However, 
// elements of rows are stored contiguously, so that rows[x][-1] == 
// rows [x-1][ny-1].  Also, rows[0][-1]== -1.
// Dimensioning this MY etc instead of NY uses more memory, but makes the
// program easier to debug.

#ifdef STATIC
run runs[mruns];   // runs, sorted by x,y.
component components[mcomps];
int rows1[mx*my+1]; 
int *(rows[mx]);   
char crow[mz+1];          // The current row, before runification.
#else
run *runs;
component *components;
int *rows1, **rows, *crow;
#endif

// istream& datafile= cin;
short int compsizehist[msize];    // Histogram of component sizes.

int *compsort;           // Which component is the i-th sorted component?
int nunitcomps;          // # volume 1 components.
int ngoodcomps;          // # interesting components

char  curinbyte;           // Current input byte, composed of 8 bit voxels.
int   curinoffset;         // Number of bits (voxels) used of current byte.
// 128 used to work; C++ is not upwards compatible.
char inmask[8]={-128,64,32,16,8,4,2,1};  // Where each bit is.
struct timeval *watchdog_tv;
struct timezone *watchdog_tz;



//  CODE

void perr(char *msg) {   
  if (errno!=0) 
    {   perror(msg); exit(2);}
}


#ifdef WATCHDOG
// WATCHDOG_PRINT Maybe print, to cerr, a message, followed by endl.
// Print if at least PERIOD=3 seconds have elapsed since the last time
// that watchdog_print printed a message.

void watchdog_print(const string msg) {
  const int period=3;
  static long lastprinttime=0;
  int ret;

  ret=gettimeofday(watchdog_tv,watchdog_tz);
  if (ret!=0) perr("gettimeofday called from watchdog_print");
  if (watchdog_tv->tv_sec-lastprinttime > period) {
    cerr << msg << endl;
    lastprinttime = watchdog_tv->tv_sec;
  }
}
#endif



// MAX   (There must be a lib with this!)

int max(const int a, const int b) {
  return (a>b) ? a : b ;
}


// MIN

int min(const int a, const int b) {
  return (a<b) ? a : b;
}



// PTIME Write cumulative CPU process time to cerr.

void ptime(const char *const msg) {
  float t= ((float)clock())/CLOCKS_PER_SEC;
  cerr << "Cumulative CPU time thru " << msg << "= " << t << endl;
}

// my exit provides a convenient place to trap exits.

void myexit(int i) {
  exit(i);
}

#ifdef FALSE

// Convert a string to an int; the builtin fails.

int myatoifails(const char *const s) {
  int i;
  istrstream f(s);
  if (f.fail()) die("strstream failed at start of myatoi");
  f >> i;
  if (f.fail()) die("strstream failed after read in myatoi");
  return i;
}

int myatoifailsalso(const char *const s) {
  int i;
  sscanf(s,"%d",&i);
  return i;
}
#endif





void initialize(const int argc, const char** const argv)
{
  int i,x;

  print_env();

  if (argc<5) {
    cerr << "argc=" << argc << endl;
    die("CONNECT requires at least 4 arguments; see the documentation.");
    }

  nx = atoi(argv[1]);
  ny = atoi(argv[2]);
  nz = atoi(argv[3]);
  i = atoi(argv[4]);
  if (argc>=6) key = atoi(argv[5]);
  else key = 0;

  cerr << "nx=" << nx << ", ny=" << ny << ", nz="
       << nz << ", conn=" << i << ", key=" << key << endl;

  //#ifndef STATIC
  mx = nx;
  my = ny;
  mz = nz;
  mruns = nx*ny*nz/10;    // a guess.
  mruns = 21000000;   // for the specific large case.
  mcomps = mruns/10;      // another guess.
  mcomps = 5000000;   // for the specific large case.
  NEWA(rows, int*, mx);
  NEWA(rows1, int, mx*my+1);
  NEWA(crow, int, mz+1);
  NEWA(runs, run, mruns);
  NEWA(components, component, mcomps);
  //#endif

  if (ny<=0 || ny > my || nz<=0 || nz>mz || nx<=0 || nx>mx)
    { 
      cerr << "mx=" << mx << ", my=" << my << ", mz=" << mz << endl;
      die("Illegal ny, nz, or nx"); 
    }

  if (i==6) conn26 = 0;
  else if (i==26) conn26 = 1;
  else {
    cerr << "ERROR illegal connectivity=" << i << ". Must be 6 or 26.  6 assumed." << endl;
    conn26 = 0;
  }
  debug = 1;
  nadjacent = 0;
  nsametree = 0;
  nvoxels = 0;
  nfollrptr = 0;
  ncomps = 0;
  nruns = 0;
  nunitcomps = 0;

  for (x=0;x<nx;x++)  
    rows[x]=rows1+ny*x+1;
  rows1[0]= -1;
  for (i=0;i<msize;i++)  
    compsizehist[i]=0;

  cin.unsetf(ios::skipws);    // Do not skip spaces on input.
  curinoffset = 8;            // Force a read before next voxel.

#ifdef WATCHDOG
  watchdog_tv = new struct timeval;
  watchdog_tz = new struct timezone;
#endif

}


void read_row(const int x, const int y)
{	
  // A row has fixed x, y, and variable iz.  Read one row.

  int iz;
  char *buf;

  if (debug>2) cerr << "Row, y=" << y << ", x=" << x << ':';

#ifdef WATCHDOG
  ostringstream sout;
  sout << "read_row, x=" << x << ", y=" << y << ends;
  watchdog_print(sout.str());
#endif

  //  if ((y%300)==0) cerr <<  "read_row, x=" << x << ", y=" << y << endl;

  for (iz=0; iz<nz; iz++)  
    {
      if (cin.eof()) 
	{ 
	  cerr << "y="<<y<<", iz="<<iz<<", x="<<x << endl;
	  die("EOF before reading.");
	}
      if (curinoffset>7) {
	cin >> curinbyte;
	if (cin.fail()) 
	  { 
	    cerr << "y="<<y<<", iz="<<iz<<", x="<<x << endl;
	    die("Input failed");
	  }
	// why does this exist? 	perr("After input");
	curinoffset = 0;
      }
      //      cin>>crow[iz];
      crow[iz] = ((curinbyte & inmask[curinoffset++])?'1':'0');
      if (debug>2) cerr << crow[iz];
    }
  if (debug>2) cerr << endl;
  crow[nz]='1';  // Insert a known termination.
  //  cerr << "After input" << endl;
}

void runnify_row(const int x, const int y)
{	//   Runnify runs of '0' voxels in current row.
  int iz, jz, area, vol;

  for (iz=0; iz<nz; iz++)
    {
      if (crow[iz] != '0') continue;
      // Starting a run.
      for (jz=iz+1; crow[jz]=='0';jz++);
      if (nruns>=mruns) {
#ifdef STATIC
	cerr << "ERROR: mruns=" << mruns << " is too small.  x=" << x
	     << ", y=" << y << endl;
	myexit(4);
#else
	cerr << "mruns=" << mruns << " is too small.  x=" << x
	     << ", y=" << y << endl;
	double extra=0.1;
	int newmruns = static_cast<int>(mruns * ((1+extra)*nx /  (x+0.) - extra));
	cerr << "Reallocating with mruns=" << newmruns << endl;
        run *runs2;
	NEWA(runs2, run, newmruns);
	for (int i=0; i<nruns; i++) runs2[i]=runs[i];
	delete[] runs;
	runs = runs2;
#endif
      }
      runs[nruns].x = x;
      runs[nruns].y = y;
      runs[nruns].zlo = iz;
      runs[nruns].zhi = jz-1;
      area = 4*(jz-iz)+2;
      vol = jz-iz;
      runs[nruns].root = -area;
      if (debug>1) cerr << "Run #" << nruns << ": y=" << y << ", z=(" << iz << ',' 
			<< jz-1	<< "), x=" << x << endl;
      nruns++;
      nvoxels +=  vol;
      iz=jz;
    }
  rows[x][y]=nruns-1;
}


//  FIND_ROOT_OF_RUN: Find the root of this run.  Return a nonnegative
//  number.

int find_root_of_run(const int run)	
{
  int r=run;
  nfindroot++;
  while (runs[r].root>=0) {
    r=runs[r].root;
    nfollrptr++;
    lfindroot++;
    if (r>=nruns) die("Illegal root in find_root_of_run");
  }
  return r;
}


// Make this run and its ancestors, down to its root, point to newroot.

void update_root(const int run, const int newroot)
{
  int r=run;
  int i;
  nupdateroot++;
  while (runs[r].root>=0) {	
    i=runs[r].root;
    runs[r].root=newroot;
    r=i;
    nfollrptr++;
    lupdateroot++;
  }
  runs[r].root=newroot;
}


// TEST_AND_PROCESS_ADJACENCY


void test_and_process_adjacency(const int run1, const int run2, const int diag)
{
  int lap1 = runs[run1].zhi - runs[run2].zlo;
  int lap2 = runs[run2].zhi - runs[run1].zlo;
  // conn26 is 0 (6-connectivity) or 1 (26-connectivity).
  if (lap1 < -conn26 || lap2 < -conn26) {
    nnonoverlap++;
    return;
  }
  int lap = min(runs[run1].zhi, runs[run2].zhi) - 
    max(runs[run1].zlo, runs[run2].zlo) +1;
  nadjacent++;
  if (debug>1) cerr << "Runs " << run1 << " and " << run2
		    << " are adjacent." << endl;
  int root1=find_root_of_run(run1);
  int root2=find_root_of_run(run2);
  if (debug>1) cerr << "Roots are " << root1 << ' ' << root2 << endl;

  if (root1==root2) {
    int area;
    if (diag) area  = -runs[root1].root;
    else area  = -runs[root1].root - 2*lap;
    if (area<0) {
      cerr << "Illegal area=" << area
	   << " in test_and_process_adjacency." << endl;
      area = 0;
      myexit(4);
    }
    update_root(run1,root1);
    runs[root1].root = -area;    // preventing an infinite loop
    update_root(run2,root1);
    runs[root1].root = -area;
    nsametree++;
  } else {
    int root=max(root1,root2);
    // Make root to be the root of the combined tree.  This maintains an
    // invariant that a run's father is higher numbered than it.
    int area1 = -runs[root1].root;   // surface area
    int area2 = -runs[root2].root;
    int area;
    if (diag) area = area1+area2;
    else area = area1+area2 - 2*lap;

    // Update root even if the root of this run is already the new root, 
    // since this compresses the tree.  If the root is already the new
    // root, then this will leave the root's root positive, which is an
    // infinite loop.  However, that is fixed by replacing the root with
    // the negative of the size.

    update_root(run1,root);
    update_root(run2,root);

    runs[root].root = -area;
    if (area<0) die("Illegal area");
  }
}



// CONNECT_ROW: Connect the runs in row (x1,y1) to the runs in row
// (x2,y2), if that row exists.  Checking here means that the caller
// needn't check.

void connect_row(const int x1, const int y1, const int x2, const int y2, const int diag)
{
  if (y1<0 || y1>=ny || x1<0 || x1>=nx || y2<0 || y2>=ny || x2<0 
      || x2>=nx) return;
  // In the following, the 2nd subscript of rows might be -1.  This is
  // OK since rows is mapped to the 1-D rows1.
  const int row1lorun=rows[x1][y1-1]+1;
  const int row1hirun=rows[x1][y1];
  const int row2lorun=rows[x2][y2-1]+1;
  const int row2hirun=rows[x2][y2];
  if (row1lorun>row1hirun || row2lorun>row2hirun) return;
  int row1currun=row1lorun;
  int row2currun=row2lorun;
 
  for(;;)
    {
      test_and_process_adjacency(row1currun, row2currun, diag);

      // In the row whose current run has the lower zhi, go to the next run.
      if (runs[row1currun].zhi<runs[row2currun].zhi) 
	{	if (++row1currun > row1hirun) break; }
      else 
	{ if (++row2currun > row2hirun) break; }
    }
}

void process_one_row(const int x, const int y)
{
  read_row(x, y);
#ifdef TIMEINPUTS
  return;
#endif
  runnify_row(x, y);
  connect_row(x, y, x, y-1, 0);		// to left
  connect_row(x, y, x-1, y, 0);		// to below
  if (conn26) {
    connect_row(x, y, x-1, y-1, 1); // to below and left.
    connect_row(x, y, x-1, y+1, 1); // to below and right.
  }
}

void process_data()
{
  int x, y;
  for (x=0;x<nx;x++)
    for(y=0;y<ny;y++)
      process_one_row(x, y);
}



void check_tree_depth_and_flatten()
{
  int irun, root, rootroot;
  for (irun=nruns-1; irun>=0; irun--) {
    const int root = runs[irun].root;
    if (root < 0) continue;
    if (root <= irun) {
      cerr << "ERROR: run[irun=" << irun << "].root=" << root 
	   << " should have root>irun." << endl;
    }
    const int rootroot = runs[root].root;
    if (rootroot < 0) continue; 
    const int root3 = runs[rootroot].root;
    if (root3<0) {
      runs[irun].root = rootroot;
      continue;
    }
    cerr << "ERROR: run[" << irun << "] is more than 2 from its root." 
	 << "run.root=" << root << ", root root=" << rootroot << endl;
      }
}

void assemble_components()
{
  int irun, root, comp;
  for (irun=nruns-1; irun>=0; irun--) {
    root = runs[irun].root;
    if (root<0) {
      if (ncomps>=mcomps) {
	cerr << "ERROR: mcomps=" << mcomps << " is too small.  irun=" 
	     << irun << ", nruns=" << nruns << endl;
	myexit(3);
      }
      components[ncomps].area = -root;
      components[ncomps].runs = irun;
      components[ncomps].nruns = 1;
      runs[irun].root = -ncomps;
      ncomps++;
    } else {
      comp = -runs[root].root;
      if (comp >= ncomps) die("illegal root");
      components[comp].nruns++;
      runs[irun].root = components[comp].runs;
      components[comp].runs = irun;
    }
  }
}

int calc_component_volume(int icomp)
{
  int irun;
  int cnrun = components[icomp].nruns;
  int thisrun = components[icomp].runs;
  int v=0;

  for (irun=0; irun<cnrun; irun++) {
    v = v - runs[thisrun].zlo + runs[thisrun].zhi + 1;
    thisrun = runs[thisrun].root;
    }
  return v;
}

int compar(const void *ii, const void *jj)
{
  const int *i = (int *) ii;
  const int *j = (int *) jj;
  if (*i<0 || *i>=ncomps || *j<0 || *j>=ncomps) {
    cerr << "ERROR: illegal args to compar.  *i= " << *i << ", *j=" << *j << endl;
    exit(10);
  }
  const int ivol=components[*i].vol;
  const int jvol=components[*j].vol;
  if (ivol<jvol) return 1;          // reverse sort
  else if (ivol==jvol) return 0;
  else return -1;
}


int sort_components()
{
  int icomp, volume;

  compsort = new int[ncomps];
  ngoodcomps = 0;
  for (icomp=0; icomp<ncomps; icomp++) {
    int &volume = components[icomp].vol;
    volume = calc_component_volume(icomp);
    if (volume > 10000000) {
      cerr << "Component #" << icomp << " with volume " << volume 
	   << " is too large to write." << endl; 
      continue;
    }
    if (volume <=1) {
      nunitcomps++;
      continue;
    }
    compsort[ngoodcomps++] = icomp;
  }
  qsort(compsort, ngoodcomps, sizeof(int), compar);
}


void write_components()
{
  int icomp, irun, cnrun;
  int volume;
  int area;
  int thisrun;
  int i;

  cout << "Connected components (see http://www.ecse.rpi.edu/Homepages/wrf/research/connect/)" << endl;
  cout << "NX NY NZ connectivity #voxels #runs #components #comps_written:" << endl;
  cout << "For each component: #runs, vol, area, (x, y, zl, zl)+" << endl;
  cout << nx << ' ' << ny << ' ' << nz 
       << (conn26?" 26 ":" 6 ") << nvoxels << ' ' << nruns << ' ' << ncomps << ' ' 
       << ngoodcomps <<  endl;

  for (icomp=0; icomp<ngoodcomps; icomp++) {

#ifdef WATCHDOG
    ostringstream sout;
    sout << "write_components, icomp=" << icomp << ends;
    watchdog_print(sout.str());
#endif

    component &thiscomp = components[compsort[icomp]];
    cnrun = thiscomp.nruns;
    volume = thiscomp.vol;
    area = thiscomp.area;
    cout << cnrun << ' ' << volume << ' ' << area;

    thisrun = thiscomp.runs; 
    for (i=0; i<cnrun; i++) {
      if (thisrun<0 || thisrun>=nruns) {
	cerr << "ERROR for icomp=" << icomp << ", cnrun=" << cnrun << ", but there were only " << i << " runs." << endl;
      }
      cout << '\t' << runs[thisrun].x << ' ' << runs[thisrun].y << ' ' 
	   << runs[thisrun].zlo << ' ' << runs[thisrun].zhi;
      thisrun = runs[thisrun].root;
    }
    cout << endl;
  }
  if (nunitcomps>0) {
    cerr << "# volume=1 components not written: " << nunitcomps << endl;
  }
}


void write_ncomponents()
{
  cout << ncomps << endl;
}



// Print_Results   Print results to cerr.

void print_results(ostream &c)
{
  const int nu=ny*nz*nx;
  int i, s1, s2, s3, s;
  const float frac= ((float) nvoxels) / (float) nu;
  const float size= ((float) nvoxels) / (float) nruns;

  c << "Universe has " << nx << '*' << ny << '*' << nz << '=' 
    << nu << " voxels." << endl;
  c << "Type of connectivity: " << (conn26 ? "26" : "6") << endl;
  c << "# empty voxels= " << nvoxels << ", fraction= " << frac << endl;
  c << "# runs= " << nruns << ", average size=" << size << endl;
  //  c << "Histogram of component sizes:" << endl;
  for (i=0; i<msize; i++)
    if (compsizehist[i]>0) c << i << ": " << compsizehist[i] << endl;
  c << "# components= " << ncomps << ", average # runs/component=" 
    << nruns/ (float)ncomps << endl;
  c << "# times a pair of possibly adjacent runs tested, but they didn't" << endl;
  c << "overlap= " << nnonoverlap << endl;
  c << "# times a pair of adjacent runs processed= " << nadjacent << endl;
  c << "# times two adjacent runs were already in the same component= " << nsametree << endl;
  c << "# pointers followed when processing adjacent runs= " << nfollrptr << endl;
  c << "Average # pointers followed per run= " << nfollrptr/(float)nruns << endl;
  c << "# times find_root_of_run called= " << nfindroot 
    << ", average # pointers followed per call= " << lfindroot / (float) nfindroot << endl;
  c << "# times update_root called= " << nupdateroot 
    << ", average # pointers followed per call= " << lupdateroot / (float) nupdateroot << endl;
  c << "# times find_roots finds a run not pointing directly to its root=" 
    << nfindrootslink << endl; 
  c << "# bytes used by one run= " << sizeof(run) << ", total run storage= " << (s1=sizeof(runs)) << endl;
  c << "# bytes used by one component= " << sizeof(component) << ", total component storage= " << (s2=sizeof(components)) << endl;
  c << "rows1 storage= " << (s3=sizeof(rows1)) <<endl;
  s= s1+s2+s3;
  c << "Total storage used by large arrays= " << s << endl;
}


main(const int argc, const char **const argv)
{
  initialize(argc, argv);
  ptime( "initialization");
  process_data();
  ptime( "processing planes");
  check_tree_depth_and_flatten();
  ptime( "check_tree_depth");
  assemble_components();
  ptime("assemble_components");
  sort_components();
  ptime("sort_components");
  if (key==0)   write_components();
  else write_ncomponents();
  ptime( "for write_components");
  print_results(cerr);
  ptime( "end");
  cerr << endl << endl
	 << "================================================================"
	 << endl << endl; 
}

//    Local Variables:
//    eval: (progn (font-lock-mode) )
//    End:
